from datetime import date

from django.db import models
from django.contrib.auth.models import User
from django.db.models import Max, Min

class CoreObject(models.Model):
	name = models.CharField(max_length = 100)
	description = models.TextField(blank = True, default = '')

	def __str__(self):
		return self.name

	def get_history(self):
		return self.history_set.filter(is_note = False).order_by('-date')

	def get_newest_history(self):
		return self.history_set.all().aggregate(Max('date'))['date__max']

	def get_oldest_history(self):
		return self.history_set.filter(is_note = False).aggregate(Min('date'))['date__min']

	def get_notes(self):
		return self.note_set.order_by('date')

# Project
class Project(CoreObject):
	prefix = models.CharField(max_length = 10, unique = True)

	def __str__(self):
		return self.prefix

	def get_build_number(self):
		build_num = self.build_set.all().aggregate(Max('number'))['number__max']
		if build_num:
			return build_num
		else:
			return 0

	def get_objective_number(self):
		obj_num = self.objective_set.all().aggregate(Max('number'))['number__max']
		if obj_num:
			return obj_num
		else:
			return 0

	def get_builds_by_date(self):
		return self.build_set.order_by('-date', '-number')

	def get_releases_by_date(self):
		return self.release_set.order_by('-release_date', '-name')

	def get_total_releases(self):
		return self.release_set.count()

	def get_total_features(self):
		return self.objective_set.filter(type = Objective.FEATURE).count()

	def get_closed_features(self):
		return self.objective_set.filter(type = Objective.FEATURE, status__in = [Objective.CLOSED, Objective.DEPRECATED]).count()

	def get_open_issues(self):
		return self.objective_set.filter(type = Objective.ISSUE, status__in = [Objective.OPEN, Objective.DEV, Objective.TEST]).count()

	def get_open_bugs(self):
		return self.objective_set.filter(type = Objective.BUG, status__in = [Objective.OPEN, Objective.DEV, Objective.TEST]).count()

	def get_users(self):
		return User.objects.all()

	def get_recent_activity(self):
		return History.objects.filter(on_project = self).order_by('-date')[:20]

	def get_all_activity(self):
		return History.objects.filter(on_project = self).order_by('-date')

	def get_recent_objectives(self):
		return self.objective_set.all().order_by('-created_date')[:10]

	def get_recent_builds(self):
		return self.build_set.order_by('-date', '-number')[:5]

# Release
class Release(CoreObject):
	release_date = models.DateField(default = date.today())
	from_project = models.ForeignKey(Project, on_delete=models.CASCADE)

	def __str__(self):
		return str(self.from_project) + '  ' + self.name

	def get_builds_by_date(self):
		return self.build_set.order_by('date', 'number')

# Build
class Build(CoreObject):
	number = models.IntegerField()
	date = models.DateField(default = date.today())
	from_project = models.ForeignKey(Project, on_delete=models.CASCADE)
	from_release = models.ForeignKey(Release, null = True, blank = True, on_delete=models.CASCADE)

	def __str__(self):
		return str(self.from_project) + ' Build ' + str(self.number)

# Objective
class Objective(CoreObject):
	FEATURE = 'Feature'
	ISSUE = 'Issue'
	BUG = 'Bug'
	OBJECTIVES = ((FEATURE, FEATURE),
		(ISSUE, ISSUE),
		(BUG, BUG),
		)

	OPEN = 'Open'
	DEV = 'In Development'
	TEST = 'In Testing'
	CLOSED = 'Closed'
	DEPRECATED = 'Deprecated'
	OBJ_STATUS = ((OPEN, OPEN),
		(DEV, DEV),
		(TEST, TEST),
		(CLOSED, CLOSED),
		(DEPRECATED, DEPRECATED),
		)

	from_project = models.ForeignKey(Project, on_delete=models.CASCADE)
	number = models.IntegerField()
	type = models.CharField(max_length = 10, choices = OBJECTIVES, default = ISSUE)
	status = models.CharField(max_length = 15, choices = OBJ_STATUS, default = OPEN)
	id_release = models.ForeignKey(Release, null = True, blank = True, on_delete=models.CASCADE, related_name = 'id_releases')
	fixed_release = models.ForeignKey(Release, null = True, blank = True, on_delete=models.CASCADE, related_name = 'fixed_releases')
	assigned_to = models.ForeignKey(User, null = True, blank = True, on_delete=models.CASCADE)
	created_date = models.DateTimeField(auto_now_add = True)

	def __str__(self):
		return self.from_project.prefix + '-' + str(self.number) + ': ' + self.name

	def status_types(self):
		return [self.OPEN, self.DEV, self.TEST, self.CLOSED, self.DEPRECATED]

	def objective_types(self):
		return [self.FEATURE, self.ISSUE, self.BUG]

	def get_users(self):
		return User.objects.all()

	def is_status_open(self):
		return self.status in [self.OPEN, self.DEV, self.TEST]

	def is_status_closed(self):
		return self.status in [self.CLOSED, self.DEPRECATED]

# Notes
class Note(models.Model):
	author = models.ForeignKey(User)
	text = models.CharField(max_length = 500)
	date = models.DateTimeField(auto_now_add = True)
	object = models.ForeignKey(CoreObject, on_delete=models.CASCADE)

	def __str__(self):
		return str(self.object) + ' - ' + str(self.author)

# History
class History(models.Model):
	user = models.ForeignKey(User)
	text = models.CharField(max_length = 255)
	date = models.DateTimeField(auto_now_add = True)
	object = models.ForeignKey(CoreObject, on_delete=models.CASCADE)
	is_note = models.BooleanField(default = False)
	on_project = models.ForeignKey(Project, on_delete=models.CASCADE, related_name = 'on_project')

	def __str__(self):
		return str(self.object) + ' - ' + str(self.user)

# # Test Case
# class TestCase(CoreModel):
# 	number = models.IntegerField()

# # Test Step
# class TestStep(CoreModel):
# 	test_case = models.ForeignKey(TestCase, on_delete=models.CASCADE)
# 	number = models.IntegerField()
# 	action = models.CharField(max_length = 255, blank = True)
# 	expected_result = models.CharField(max_length = 255, blank = True)

# # Test Execution
# class TestExecution(models.Model):
# 	test_case = models.ForeignKey(TestCase, on_delete=models.CASCADE)
# 	date = models.DateField(auto_now = True)
# 	build = models.ForeignKey(Build, on_delete=models.CASCADE)
# 	tester = models.ForeignKey(User)

# # Execution Step
# class ExecutionStep(models.Model)
# 	PASSED = 'Passed'
# 	FAILED = 'Failed'
# 	BLOCKED = 'Blocked'
# 	TEST_STATUS((PASSED, PASSED),
# 		(FAILED, FAILED),
# 		(BLOCKED, BLOCKED),
# 		)
# 	test_step = models.ForeignKey(TestStep, on_delete=models.CASCADE)
# 	actual_result = models.CharField(max_length = 255, blank = True)
# 	status = models.CharField(max_length = 10, choices = OBJECTIVES, null = True)

# # Test Plan
# class TestPlan(CoreObject):
# 	tests = models.ManyToManyField(TestCase)
