import logging
from datetime import datetime

from django.apps import apps
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseServerError
from django.core.urlresolvers import reverse
from django.shortcuts import render
from django.views import View
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.models import User
from django.contrib.auth.mixins import LoginRequiredMixin

from projectcx.models import CoreObject, Project, Release, Build, Objective, Note, History

# Login
def login_projectcx(request):
	logger = logging.getLogger('django.projectcx.login')

	context = {'errors': {}}
	errors = context['errors']
	logout(request)
	# process the request
	if request.method == 'POST':
		#print(request.POST)
		username = request.POST['username']
		password = request.POST['password']
		user = authenticate(username=username, password=password)
		if user:
			if user.is_active:
				login(request, user)
				return HttpResponseRedirect(reverse('projectcx:project_list'))
			else:
				errors['page_error'] = 'User has been disabled.'
		else:
			logger.warn("Invalid login details: {0}, {1}".format(username, password))
			errors['page_error'] = 'Username and/or password is incorrect.'
	return render(request, 'projectcx/login.html', context)



def live_edit(request):
	logger = logging.getLogger('django.projectcx.liveedit')
	'''
	Receives a POST request with the following parameters:
		
		modelid: the database id of the model
		modelname: the name of the model object
		fieldname: the field of the object to edit
		value: the value to store in the field

	The framework sends SUCCESS unless there is an error, which results in an HTTP 500 (server error).
	If request is not a POST, the request will receive an HTTP 405 (response not allowed).
		'''
	date_fields = ['date', 'release_date']
	id_fields = {'Build':{'from_release': Release.objects},
				'Objective':{'id_release': Release.objects,
							'fixed_release': Release.objects,
							'assigned_to': User.objects},
				}

	if request.method != 'POST':
		logger.exception('LiveEdit request is not a POST request')
		return HttpResponseNotAllowed(['POST'])
	# parse the request
	try:
		modelid = request.POST['modelid']
		modelname = request.POST['modelname']
		fieldname = request.POST['fieldname']
		value = request.POST['value']
		if fieldname in date_fields:
			# convert to date object
			value = datetime.strptime(request.POST['value'], '%Y-%m-%d')
		# foreign key fields with an id of -1 should be set to None
		if modelname in id_fields.keys():
			if fieldname in id_fields[modelname].keys():
				model_id = int(value)
				model_manager = id_fields[modelname][fieldname]
				if model_id < 0:
					value = None
				else:
					value = model_manager.get(pk=model_id)
	except:
		logger.exception('LiveEdit POST is invalid')
		return HttpResponseServerError(reason = "POST request is invalid.")
	# find the model object manager
	try:
		model = apps.get_model('projectcx', modelname)
	except:
		logger.exception('LiveEdit model is invalid')
		return HttpResponseServerError(reason = "Model is not valid.")
	# find the field
	try:
		field = model._meta.get_field(fieldname)
	except:
		logger.exception('LiveEdit field is invalid')
		return HttpResponseServerError(reason = "Field is not valid.")

	try:
		# if we've marked this as a new update
		# need to create a new ItemUpdate object and populate it
		if 'req_vals' in request.POST.keys():
			required_values = json.loads(request.POST['req_vals'])
			the_model = model.objects.get(pk = modelid)
		else:
			the_model = model.objects.get(pk = modelid)
		# update the field
		setattr(the_model, fieldname, value)
		# save the model
		the_model.save()
		obj = CoreObject.objects.get(pk=the_model.id)
		text = str(request.user.get_full_name()) + ' altered ' + str(the_model) + ':  ' + fieldname + ' = ' + str(value)
		project = None
		if modelname == 'Project':
			project = the_model
		else:
			try:
				project = the_model.from_project
			except:
				logger.exception('Object does not have a from_project attribute!')
		History.objects.create(user = request.user, text = text, object = obj, on_project = project)
	except:
		logger.exception('LiveEdit model save failed')
		return HttpResponseServerError(reason = "Model could not be saved.")
	# return SUCCESS
	logger.info('Successful LiveEdit request')
	return HttpResponse('SUCCESS')


def create_note(request, project_id):
	logger = logging.getLogger('django.projectcx.create_note')
	project_id = int(project_id)
	if request.method != 'POST':
		logger.exception('Create Note request is not a POST request')
		return HttpResponseNotAllowed(['POST'])
	obj_id = int(request.POST['object_id'])
	text = request.POST['new_note']

	project = Project.objects.get(pk=project_id)
	obj = CoreObject.objects.get(pk=obj_id)
	note = Note.objects.create(author = request.user, text = text, object = obj)
	history = str(request.user.get_full_name()) + ' wrote a note on ' + str(obj)
	History.objects.create(user = request.user, text = history, object = obj, is_note = True, on_project = project)
	return HttpResponseRedirect(reverse('projectcx:read_objective', kwargs = {'project_id': project_id, 'objective_id': obj_id}))


class CoreView(LoginRequiredMixin, View):
	login_url = 'projectcx:login'

	def get_context_data(self, *args, **kwargs):
		context = {}
		context['errors'] = {}
		return context

	def get(self, request, *args, **kwargs):
		return self.do_get(request, self.get_context_data())

	def do_get(self, request, context):
		return HttpResponse('ERROR: do_get not implemented!')

	def post(self, request, *args, **kwargs):
		context = super(CoreView, self).get_context_data(**request.POST)
		return self.do_post(request, context)

	def do_post(self, request, context):
		return HttpResponse('ERROR: do_post not implemented!')

# Logout
class LogoutView(CoreView):

	def get(self, request):
		logout(request)
		return HttpResponseRedirect(reverse('projectcx:login'))

# View Users
class UserList(CoreView):

	def do_get(self, request, context):
		context['user_list'] = User.objects.all().order_by('username')
		return render(request, 'projectcx/user_list.html', context)


# View Projects
class ProjectList(CoreView):

	def do_get(self, request, context):
		context['project_list'] = Project.objects.all()
		return render(request, 'projectcx/project_list.html', context)

class ProjectView(CoreView):

	def get_context_data(self, *args, **kwargs):
		context = super(ProjectView, self).get_context_data()
		context['project'] = Project.objects.get(pk = self.kwargs['project_id'])
		return context

# Project Home
class ProjectHome(ProjectView):

	def do_get(self, request, context):
		return render(request, 'projectcx/project_home.html', context)

# Releases
class Releases(ProjectView):

	def do_get(self, request, context):
		context['releases'] = Release.objects.filter(from_project = self.kwargs['project_id']).order_by('-release_date', '-name')
		context['other_builds'] = Build.objects.filter(from_project = self.kwargs['project_id'], from_release = None).order_by('date')
		return render(request, 'projectcx/releases.html', context)

# Features
class Features(ProjectView):

	def do_get(self, request, context):
		context['page_title'] = "Features"
		context['objective_type'] = "Feature"
		context['objectives'] = Objective.objects.filter(from_project = self.kwargs['project_id'], type = Objective.FEATURE).order_by('number')
		return render(request, 'projectcx/objectives.html', context)

# Issues & Bugs
class Issues(ProjectView):

	def do_get(self, request, context):
		context['page_title'] = "Issues & Bugs"
		context['objective_type'] = "Issue"
		context['objectives'] = Objective.objects.filter(from_project = self.kwargs['project_id'], type__in = [Objective.ISSUE, Objective.BUG]).order_by('number')
		return render(request, 'projectcx/objectives.html', context)

# Objectives
class Objectives(ProjectView):

	def do_get(self, request, context):
		context['page_title'] = "All Objectives"
		context['objective_type'] = "Objective"
		context['objectives'] = Objective.objects.filter(from_project = self.kwargs['project_id']).order_by('number')
		return render(request, 'projectcx/objectives.html', context)

# User Objectives
class UserObjectives(CoreView):

	def do_get(self, request, context):
		context['page_title'] = "My Objectives"
		context['objective_type'] = "Objective"
		context['objectives'] = Objective.objects.filter(assigned_to = request.user).order_by('created_date')
		return render(request, 'projectcx/user_objectives.html', context)

class CoreCRUD(CoreView):
	CREATE = 'create'
	READ = 'read'
	UPDATE = 'update'
	DELETE = 'delete'

	mode = READ
	logger = logging.getLogger('django.projectcx.generic')
	context_name = None
	id_param = None
	template = None
	redirect = None

	# def get_context_data(self, *args, **kwargs):
	# 	context = super(ProjectCRUD, self).get_context_data()
	# 	context['mode'] = self.kwargs['mode']
	# 	return context

	def do_get(self, request, context):
		if self.mode == self.CREATE:
			return self.do_create(request, context, self.create(request, context) )
		elif self.mode == self.READ:
			return self.do_read(request, context, self.read(request, context) )
		elif self.mode == self.UPDATE:
			return self.do_update(request, context, self.update(request, context) )
		elif self.mode == self.DELETE:
			return self.do_delete(request, context, self.delete(request, context) )
		return HttpResponse('ERROR: mode context not recognized!')

	def do_post(self, request, context):
		if self.mode == self.CREATE:
			return self.do_create(request, context, self.create(request, context) )
		elif self.mode == self.READ:
			return self.do_read(request, context, self.read(request, context) )
		elif self.mode == self.UPDATE:
			return self.do_update(request, context, self.update(request, context) )
		elif self.mode == self.DELETE:
			return self.do_delete(request, context, self.delete(request, context) )
		return HttpResponse('ERROR: mode context not recognized!')

	def create(self, request, context):
		return None

	def do_create(self, request, context, obj):
		if not obj or not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: create not implemented correctly!')
		try:
			obj.save()
			text = str(request.user) + ' created a ' + obj.__class__.__name__
			History.objects.create(user = request.user, object = CoreObject.objects.get(pk=obj.id), text = text)
			context[self.context_name] = obj
			return render(request, self.template, context)
		except:
			self.logger.exception('do_create failed!')
			return HttpResponseRedirect(reverse(self.redirect))

	def read(self, request, context):
		return None

	def do_read(self, request, context, obj):
		if not obj or not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: read not implemented correctly!')
		try:
			context[self.context_name] = obj
			return render(request, self.template, context)
		except:
			self.logger.exception('do_read failed!')
			return HttpResponseRedirect(reverse(self.redirect))

	def update(self, request, context):
		return None

	def do_update(self, request, context, obj):
		if not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: update not implemented correctly!')
		return HttpResponseRedirect(reverse(self.redirect))

	def delete(self, request, context):
		return None

	def do_delete(self, request, context, obj):
		if not obj or not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: delete not implemented correctly!')
		try:
			obj.delete()
		except:
			self.logger.exception('do_delete failed!')
		return HttpResponseRedirect(reverse(self.redirect))


class ProjectCRUD(ProjectView, CoreCRUD):

	def do_create(self, request, context, obj):
		if not obj or not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: create not implemented correctly!')
		try:
			obj.save()
			text = str(request.user) + ' created a ' + obj.__class__.__name__
			History.objects.create(user = request.user, object = CoreObject.objects.get(pk=obj.id), text = text, on_project = context['project'])
			context[self.context_name] = obj
			return render(request, self.template, context)
		except:
			self.logger.exception('do_create failed!')
			return HttpResponseRedirect(reverse(self.redirect, kwargs = {'project_id': context['project'].id}))

	def do_read(self, request, context, obj):
		if not obj or not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: read not implemented correctly!')
		try:
			context[self.context_name] = obj
			return render(request, self.template, context)
		except:
			self.logger.exception('do_read failed!')
			return HttpResponseRedirect(reverse(self.redirect, kwargs = {'project_id': context['project'].id}))

	def do_update(self, request, context, obj):
		if not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: update not implemented correctly!')
		return HttpResponseRedirect(reverse(self.redirect, kwargs = {'project_id': context['project'].id}))

	def do_delete(self, request, context, obj):
		if not obj or not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: delete not implemented correctly!')
		try:
			obj.delete()
		except:
			self.logger.exception('do_delete failed!')
		return HttpResponseRedirect(reverse(self.redirect, kwargs = {'project_id': context['project'].id}))


# User CRUD
class UserCRUD(CoreCRUD):
	logger = logging.getLogger('django.projectcx.user')
	context_name = 'myuser'
	id_param = 'user_id'
	template = 'projectcx/edit_user.html'
	redirect = 'projectcx:user_list'

	def create(self, request, context):
		return User(username = 'NewUser')

	def read(self, request, context):
		return User.objects.get(pk = self.kwargs[self.id_param])

	def delete(self, request, context):
		return User.objects.get(pk = self.kwargs[self.id_param])


# Project Main CRUD
class ProjectMainCRUD(CoreCRUD):
	logger = logging.getLogger('django.projectcx.project')
	context_name = 'project'
	id_param = 'project_id'
	template = 'projectcx/edit_project.html'
	redirect = 'projectcx:project_list'

	def create(self, request, context):
		num_proj = Project.objects.count() + 1
		return Project(name = 'New Project', prefix = 'PROJ-' + str(num_proj) )

	def do_create(self, request, context, obj):
		if not obj or not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: create not implemented correctly!')
		try:
			obj.save()
			text = str(request.user) + ' created a ' + obj.__class__.__name__
			History.objects.create(user = request.user, object = CoreObject.objects.get(pk=obj.id), text = text, on_project = obj)
			context[self.context_name] = obj
			return render(request, self.template, context)
		except:
			self.logger.exception('do_create failed!')
			return HttpResponseRedirect(reverse(self.redirect))

	def read(self, request, context):
		return Project.objects.get(pk = self.kwargs[self.id_param])

	def update(self, request, context):
		return Project.objects.get(pk = self.kwargs[self.id_param])

	def do_update(self, request, context, obj):
		if not self.context_name or not self.id_param or not self.template or not self.redirect:
			return HttpResponse('ERROR: update not implemented correctly!')
		return HttpResponseRedirect(reverse('projectcx:project_home', kwargs = {'project_id': obj.id}))

	def delete(self, request, context):
		return Project.objects.get(pk = self.kwargs[self.id_param])


# Release CRUD
class ReleaseCRUD(ProjectCRUD):
	logger = logging.getLogger('django.projectcx.release')
	context_name = 'release'
	id_param = 'release_id'
	template = 'projectcx/edit_release.html'
	redirect = 'projectcx:releases'

	def create(self, request, context):
		return Release(name = 'New Release', from_project = context['project'])

	def read(self, request, context):
		return Release.objects.get(pk = self.kwargs[self.id_param])

	def delete(self, request, context):
		return Release.objects.get(pk = self.kwargs[self.id_param])

# Build CRUD
class BuildCRUD(ProjectCRUD):
	logger = logging.getLogger('django.projectcx.build')
	context_name = 'build'
	id_param = 'build_id'
	template = 'projectcx/edit_build.html'
	redirect = 'projectcx:releases'

	def create(self, request, context):
		number = context['project'].get_build_number() + 1
		return Build(name = 'New Build', number = number, from_project = context['project'])

	def read(self, request, context):
		return Build.objects.get(pk = self.kwargs[self.id_param])

	def delete(self, request, context):
		return Build.objects.get(pk = self.kwargs[self.id_param])

# Objective CRUD
class ObjectiveCRUD(ProjectCRUD):
	logger = logging.getLogger('django.projectcx.build')
	context_name = 'objective'
	id_param = 'objective_id'
	template = 'projectcx/edit_objective.html'
	redirect = 'projectcx:objectives'

	def create(self, request, context):
		number = context['project'].get_objective_number() + 1
		return Objective(name = 'New Objective', number = number, from_project = context['project'])

	def read(self, request, context):
		return Objective.objects.get(pk = self.kwargs[self.id_param])

	def delete(self, request, context):
		return Objective.objects.get(pk = self.kwargs[self.id_param])




